// ========================================================================
// matrix3x3.h
//
// R^(3x3)
// ========================================================================

#ifndef	__MATRIX3X3_H
#define	__MATRIX3X3_H

// | e11 e12 e13 |
// | e21 e22 e23 |
// | e31 232 e33 |

// | v1 v2 v3|


#include <algorithm>
#include <math.h>
#include "vector3.h"
#include "vector4.h"

#undef  _MATRIX_ROWMAX
#define	_MATRIX_ROWMAX	3

//
//
class matrix3x3
{
friend vector3;
friend vector4;
friend class matrix4x4;

private:
	vector3 e[3];

public:
	matrix3x3() {}
	matrix3x3(VAR   e11,VAR   e12,VAR   e13,
	       VAR   e21,VAR   e22,VAR   e23,
		   VAR   e31,VAR   e32,VAR   e33 );
	matrix3x3(const matrix3x3& a);
	matrix3x3(const VAR   a);
	matrix3x3(vector3& e1,vector3& e2,vector3& e3){
		e[0]=e1;e[1]=e2;e[2]=e3;
	}
	matrix3x3(VAR   e1,VAR   e2,VAR   e3){
		e[0]=vector3(e1,0,0); e[1]=vector3(0,e2,0); e[2]=vector3(0,0,e3);
	}

	~matrix3x3() {}

	void clear();

	// one term operator
	vector3 operator[](int n) const { return e[n]; }
	vector3 row(int n) { return e[n]; }
	vector3 col(int n) { return vector3(e[0][n],e[1][n],e[2][n]); }

	VAR  & operator()(int i,int j){ return e[j][i]; }


	// two term operator
	friend matrix3x3 operator+(matrix3x3& a,matrix3x3& b);
	friend matrix3x3 operator-(matrix3x3& a,matrix3x3& b);
	friend matrix3x3 operator*(matrix3x3& a,matrix3x3& b);
	friend matrix3x3 operator*(VAR        a,matrix3x3& b);
	friend matrix3x3 operator*(matrix3x3& a,VAR        b);
	friend matrix3x3 operator/(matrix3x3& a,matrix3x3& b);
	friend matrix3x3 operator/(matrix3x3& a,VAR        b);

	friend vector3   operator*(matrix3x3& a,vector3&   v);

	matrix3x3& operator+=(matrix3x3& a);
	matrix3x3& operator-=(matrix3x3& a);
	matrix3x3& operator*=(matrix3x3& a);
	matrix3x3& operator/=(matrix3x3& a);


	//
	matrix3x3 inv();
	matrix3x3 transpose();


	//
	void setrotx(VAR   t);
	void setroty(VAR   t);
	void setrotz(VAR   t);
};

inline matrix3x3::matrix3x3(VAR   e11,VAR   e12,VAR   e13,
	       VAR   e21,VAR   e22,VAR   e23,
		   VAR   e31,VAR   e32,VAR   e33 )
{
	e[0]=vector3(e11,e21,e31);
	e[1]=vector3(e12,e22,e32);
	e[2]=vector3(e13,e23,e33);
}


inline void matrix3x3::clear()
{
	*this=0.0f;
}


inline matrix3x3::matrix3x3(const matrix3x3& a)
{
	for(int i=3-1;i>=0;i--)
		e[i]=a.e[i];
}


inline matrix3x3::matrix3x3(const VAR   a)
{
	e[0]=vector3(a,0,0);
	e[1]=vector3(0,a,0);
	e[2]=vector3(0,0,a);
}


inline matrix3x3 operator+(matrix3x3& a,matrix3x3& b)
{
	return matrix3x3(
		a.e[0]+b.e[0],a.e[1]+b.e[1],a.e[2]+b.e[2]
		);
}

inline matrix3x3 operator-(matrix3x3& a,matrix3x3& b)
{
	return matrix3x3(
		a.e[0]-b.e[0],a.e[1]-b.e[1],a.e[2]-b.e[2]
		);
}

inline matrix3x3 operator*(matrix3x3& a,matrix3x3& b)
{
	matrix3x3 c;
	for(int i=3-1;i>=0;i--)
	{
		for(int j=3-1;j>=0;j--)
		{
			VAR   f=0;
			for(int k=0;k<3;k++)
				f+=a.e[k][j]*b.e[i][k];
			c.e[i][j]=f;
		}
	}
	return c;
}

inline matrix3x3 operator*(VAR   a,matrix3x3& b){
	return matrix3x3( a*b.e[0],a*b.e[1],a*b.e[2] );
}

inline matrix3x3 operator*(matrix3x3& a,VAR   b){
	return b*a;
}

inline matrix3x3 operator/(matrix3x3& a,matrix3x3& b){
	return a*b.inv();
}

inline matrix3x3 operator/(matrix3x3& a,VAR   b){
	return a*(1.0f/b);
}

inline vector3   operator*(matrix3x3& a,vector3&   v){
	return vector3( a.col(0)*v,a.col(1)*v,a.col(2)*v );
}


inline matrix3x3 matrix3x3::inv()
{
	matrix3x3 a,b;
	a=*this;
	b=1.0f;
	for(int i=0;i<_MATRIX_ROWMAX;i++)
	{
		int u=i;
		for(int k=i+1;k<_MATRIX_ROWMAX;k++)
		{
			if( fabs(a.e[k][i])>fabs(a.e[u][i]) )
				u=k;
		}
		std::swap(a.e[u],a.e[i]);
		std::swap(b.e[u],b.e[i]);

		b.e[i]/=a.e[i][i];
		a.e[i]/=a.e[i][i];
		for(int j=0;j<_MATRIX_ROWMAX;j++)
		{
			if(i==j) continue;
			VAR   d=(a.e[j][i]/a.e[i][i]);
			a.e[j]-=a.e[i]* d;
			b.e[j]-=b.e[i]* d;
		}
	}

	return b;
}

inline matrix3x3 matrix3x3::transpose()
{
	return matrix3x3(
		e[0][0],e[0][1],e[0][2],
		e[1][0],e[1][1],e[1][2],
		e[2][0],e[2][1],e[2][2]
		);
}


inline matrix3x3& matrix3x3::operator+=(matrix3x3& a)
{
	*this=*this+a;
	return *this;
}

inline matrix3x3& matrix3x3::operator-=(matrix3x3& a)
{
	*this=*this-a;
	return *this;
}

inline matrix3x3& matrix3x3::operator*=(matrix3x3& a)
{
	*this=*this*a;
	return *this;
}

inline matrix3x3& matrix3x3::operator/=(matrix3x3& a)
{
	*this=*this/a;
	return *this;
}


inline void matrix3x3::setrotx(VAR   t)
{
	*this=1.0f;
	e[1][1]= cos(t);
	e[2][1]=-sin(t);
	e[1][2]= sin(t);
	e[2][2]= cos(t);
}

inline void matrix3x3::setroty(VAR   t)
{
	*this=1.0f;
	e[0][0]= cos(t);
	e[2][0]= sin(t);
	e[0][2]=-sin(t);
	e[2][2]= cos(t);
}

inline void matrix3x3::setrotz(VAR   t)
{
	*this=1.0f;
	e[0][0]= cos(t);
	e[1][0]=-sin(t);
	e[0][1]= sin(t);
	e[1][1]= cos(t);
}



#endif //__MATRIX3X3_H
